/*****************************************************************************
Copyright 1987, 1988, 1989, 1990, 1991 by Digital Equipment Corp., Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *
 *  CONTRIBUTORS:
 *
 *      Dick Annicchiarico
 *      Robert Chesler
 *      Dan Coutu
 *      Gene Durso
 *      Marc Evans
 *      Alan Jamison
 *      Mark Henry
 *      Ken Miller
 *
 */

#define NEED_REPLIES
#define NEED_EVENTS


#include "xtraplib.h"
#include "xtraplibproto.h"

#ifndef lint
static char RCSID[] = "$Header$";
#endif

#ifndef TRUE
# define TRUE 1L
#endif
#ifndef FALSE
# define FALSE 0L
#endif

static XETC     TC;

/*
 * This function is used to create a new XTrap context structure. The new
 * context is initialized to a hard coded default, then modified by the
 * valuemask and values passed in by the caller.
 */

#ifdef FUNCTION_PROTOS
XETC *XECreateTC(Display *dpy, CARD32 valuemask, XETCValues *value)
#else
XETC *XECreateTC(dpy, valuemask, value)  /* Returns Pointer to new TC */
    Display *dpy;                         /* Ptr to opened display */
    CARD32 valuemask;                     /* which ones to set initially */
    XETCValues *value;                   /* the values themselves */
#endif
{
    static Bool firsttime = True;
    register XETC *tc     = &TC;
    XETrapGetAvailRep rep;

    /* If this is the first time here, then initialize the default TC */
    if (firsttime == True)
    {   
        firsttime = False;
        /* The first Trap Context is the Template (default) TC */
        (void)memset(tc,0L,sizeof(tc));
        tc->eventBase             = 0x7FFFFFFFL;
        tc->errorBase             = 0x7FFFFFFFL;
        tc->values.v.max_pkt_size = 0x7FFFL;
        tc->values.v.transport    = FamilyXLib;
        BitTrue(tc->values.tc_flags, FamilyXLib);
#ifndef _XINPUT
        tc->values.v.read_transport    = FamilyXLib;
#endif
#ifdef TCPCONN
        BitTrue(tc->values.tc_flags, FamilyInternet);
#endif
#ifdef DNETCONN
        BitTrue(tc->values.tc_flags, FamilyDECnet);
#endif
#ifdef SHMEMCONN
        BitTrue(tc->values.tc_flags, FamilyShMem);
#endif
#ifdef MBXCONN
        BitTrue(tc->values.tc_flags, FamilyMbx);
#endif
#ifdef LATCONN
        BitTrue(tc->values.tc_flags, FamilyLAT);
#endif
    }

    /* Position to the end of the list */
    for (;tc->next != NULL; tc = tc->next);

    /* Allocate memory for the new context */
    if ((tc = (tc->next = (XETC *)XtMalloc(sizeof(*tc)))) == NULL)
    {   /* No memory to build TC, XtMalloc has already reported the error */
        return(NULL);
    }

    /* Use the original TC as the template to start from */
    (void)memcpy(tc,&TC,sizeof(TC));
    tc->next      = NULL;
    tc->dpy       = dpy;
    tc->xmax_size = XMaxRequestSize(tc->dpy);

    /* Initialize Extension */
    if (!XETrapQueryExtension(dpy,&(tc->eventBase),&(tc->errorBase),
        &(tc->extOpcode)))
    {
        char *params = XTrapExtName;
        int num_params = 1L;
        XtErrorMsg("CantLoadExt", "XECreateTC", "XTrapToolkitError",
            "Can't load %s extension", &params, &num_params);
        (void)XtFree(tc);
        return(NULL);
    }

    /* Allocate memory for the XLIB transport */
    if ((tc->xbuff = (BYTE *)XtMalloc(tc->xmax_size * sizeof(CARD32) +
        XETrapHeaderSize)) == NULL)
    {   /* No memory to build TC, XtMalloc has already reported the error */
        return(NULL);
    }

    /* Determine the event mask used by the extension */
    if (XEGetAvailableRequest(tc,&rep) == True)
    {
        tc->eventMask = rep.event_mask;
    }

    /* Assign the context values the caller provided */
    (void)XEChangeTC(tc, valuemask, value);

    return (tc);
}


#ifdef FUNCTION_PROTOS
static int CheckChangeBits(XETrapFlags *dest, XETrapFlags *src, INT32 bit)
#else
static int CheckChangeBits(dest,src,bit)
    XETrapFlags *dest;
    XETrapFlags *src;
    INT32      bit;
#endif
{
    int chg_flag = False;

    if (!(BitIsFalse(dest->valid,bit) && BitIsFalse(src->valid,bit)) ||
        !(BitIsTrue(dest->valid,bit) && BitIsTrue(src->valid,bit)))
    {
        BitCopy(dest->valid, src->valid, bit);
        chg_flag = True;
    }
    if (!(BitIsFalse(dest->data,bit) && BitIsFalse(src->data,bit)) ||
        !(BitIsTrue(dest->data,bit) && BitIsTrue(src->data,bit)))
    {
        BitCopy(dest->data, src->data, bit);
        chg_flag = True;
    }
    return(chg_flag);
}

/*
 * This function is called to change one or more parameters used to define
 * a context in which XTrap is or will be running.
 */
#ifdef FUNCTION_PROTOS
int XEChangeTC(XETC *tc, CARD32 mask, XETCValues *values)
#else
int XEChangeTC(tc, mask, values)
    XETC *tc;
    CARD32 mask;
    XETCValues *values;
#endif
{
    int status = True;
    register XETCValues *tval = &(tc->values);
    register int i;

    if (mask & TCStatistics)
    {   /* Statistics need changing */
        if(CheckChangeBits(&(tval->v.flags), &(values->v.flags), 
            XETrapStatistics))
        {
            tc->dirty |= TCStatistics;
        }
    }
    if (mask & TCRequests)
    {   /* Requests need changing */
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapRequest);
        for (i=0; i<256L; i++)
        {
            XETrapSetCfgFlagReq(tval, i, BitValue(values->v.flags.req,i));
        }
        tc->dirty |= TCRequests;
    }
    if (mask & TCEvents)
    {   /* Events need changing */
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapEvent);
        for (i=KeyPress; i<=MotionNotify; i++)
        {
            XETrapSetCfgFlagEvt(tval, i, BitValue(values->v.flags.event,i));
        }
        tc->dirty |= TCEvents;
    }
    if (mask & TCMaxPacket)
    {   /* MaxPacket needs changing */
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapMaxPacket);
        XETrapSetCfgMaxPktSize(tval, values->v.max_pkt_size);
        tc->dirty |= TCMaxPacket;
    }
    if (mask & TCCmdKey)
    {   /* CmdKey needs changing */
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapCmd);
        tval->v.cmd_key = values->v.cmd_key;
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapCmdKeyMod);
        tc->dirty |= TCCmdKey;
    }
    if (mask & TCTimeStamps)
    {   /* TimeStamps needs changing */
        if (CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapTimestamp))
        {
            tc->dirty |= TCTimeStamps;
        }
        BitCopy(tval->tc_flags, values->tc_flags, XETCDeltaTimes);
    }
    if (mask & TCWinXY)
    {   /* Window XY's need changing */
        if (CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapWinXY))
        {
            tc->dirty |= TCWinXY;
        }
    }
    if (mask & TCCursor)
    {   /* Window XY's need changing */
        if (CheckChangeBits(&(tval->v.flags), &(values->v.flags),XETrapCursor))
        {
            tc->dirty |= TCCursor;
        }
    }
    if (mask & TCStream)
    {   /* Stream needs changing */
        tc->dirty |= TCStream;
        XETrapSetCfgStream(tval, values->stream);
        XETrapSetCfgTransport(tval, values->v.transport);
        XETrapSetCfgAddress(tval, values->v.address);
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapTransOut);
#ifndef _XINPUT
        CheckChangeBits(&(tval->v.flags), &(values->v.flags), XETrapTransIn);
        XETrapSetCfgReadTransport(tval, values->v.read_transport);
#endif
    }
    if (XETrapGetTCFlagTrapActive(tc) || (mask & TCStream))
    {
        status = XEFlushConfig(tc);
    }
#ifdef VMS
    sys$setast(True);   /* Make sure AST's are enabled */
#endif /* VMS */
    return(status);
}


#ifdef FUNCTION_PROTOS
void XEFreeTC(XETC *tc)
#else
void XEFreeTC(tc)
    XETC *tc;
#endif
{
    register XETC *list = &TC;

    if (tc)
    {
        while(list->next != NULL)
        {
            if (list->next == tc)
            {   /* Got it, now remove it from the list */
                list->next = list->next->next;
            }
        }    
        if (tc->values.req_cb)
        {
            XtFree(tc->values.req_cb);
        }
        if (tc->values.evt_cb)
        {
            XtFree(tc->values.evt_cb);
        }
        XtFree(tc);
    }
    return;
}

/* The following are Convenience routines for setting values within
 * the Trap Context.  These are analogous to the GC's Convenience
 * Functions such as XSetState & XSetForeground
 */
#ifdef FUNCTION_PROTOS
int XETrapSetMaxPacket(XETC *tc, Bool set_flag, CARD16 size)
#else
int XETrapSetMaxPacket(tc, set_flag, size)
    XETC   *tc;
    Bool   set_flag;
    CARD16 size;
#endif
{
    XETCValues tcv;
    int status = True;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagMaxPacket(&tcv, valid, True);
    XETrapSetCfgFlagMaxPacket(&tcv, data, set_flag);
    XETrapSetCfgMaxPktSize(&tcv, size);
    status = XEChangeTC(tc, TCMaxPacket, &tcv);
    return(status);
}
#ifdef FUNCTION_PROTOS
int XETrapSetCommandKey(XETC *tc, Bool set_flag, KeySym cmd_key, Bool mod_flag)
#else
int XETrapSetCommandKey(tc, set_flag, cmd_key, mod_flag)
    XETC    *tc;
    Bool    set_flag;
    KeySym  cmd_key;
    Bool    mod_flag;
#endif
{
    XETCValues tcv;
    int status = True;
    KeyCode cmd_keycode;

    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagCmd(&tcv, valid, True);
    XETrapSetCfgFlagCmd(&tcv, data, set_flag);
    XETrapSetCfgFlagCmdKeyMod(&tcv, valid, True);
    XETrapSetCfgFlagCmdKeyMod(&tcv, data, mod_flag);
    if (!(cmd_keycode = XKeysymToKeycode(XETrapGetDpy(tc), cmd_key)))
    {
        status = False;
    }
    else
    {
        XETrapSetCfgCmdKey(&tcv, cmd_keycode);
        status = XEChangeTC(tc, TCCmdKey, &tcv);
    }
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetTimestamps(XETC *tc, Bool set_flag, Bool delta_flag)
#else
int XETrapSetTimestamps(tc, set_flag, delta_flag)
    XETC    *tc;
    Bool    set_flag;
    Bool    delta_flag;
#endif
{
    XETCValues tcv;
    int status = True;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagTimestamp(&tcv, valid, True);
    XETrapSetCfgFlagTimestamp(&tcv, data, set_flag);
    XETrapSetValFlagDeltaTimes(&tcv, delta_flag);
    status = XEChangeTC(tc, TCTimeStamps, &tcv);
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetWinXY(XETC *tc, Bool set_flag)
#else
int XETrapSetWinXY(tc, set_flag)
    XETC    *tc;
    Bool    set_flag;
#endif
{
    XETCValues tcv;
    int status = True;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagWinXY(&tcv, valid, True);
    XETrapSetCfgFlagWinXY(&tcv, data, set_flag);
    status = XEChangeTC(tc, TCWinXY, &tcv);
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetCursor(XETC *tc, Bool set_flag)
#else
int XETrapSetCursor(tc, set_flag)
    XETC    *tc;
    Bool    set_flag;
#endif
{
    XETCValues tcv;
    int status = True;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagCursor(&tcv, valid, True);
    XETrapSetCfgFlagCursor(&tcv, data, set_flag);
    status = XEChangeTC(tc, TCCursor, &tcv);
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetStatistics(XETC *tc, Bool set_flag)
#else
int XETrapSetStatistics(tc, set_flag)
    XETC    *tc;
    Bool    set_flag;
#endif
{
    XETCValues tcv;
    int status = True;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagStatistics(&tcv, valid, True);
    XETrapSetCfgFlagStatistics(&tcv, data, set_flag);
    status = XEChangeTC(tc, TCStatistics, &tcv);
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetRequests(XETC *tc, Bool set_flag, ReqFlags requests)
#else
int XETrapSetRequests(tc, set_flag, requests)
    XETC     *tc;
    Bool     set_flag;
    ReqFlags requests;
#endif
{
    XETCValues tcv;
    int status = True;
    int i;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagRequest(&tcv, valid, True);
    XETrapSetCfgFlagRequest(&tcv, data, set_flag);
    for (i=0; i<256L; i++)
    {
        XETrapSetCfgFlagReq(&tcv, i, BitValue(requests, i));
    }
    status = XEChangeTC(tc, TCRequests, &tcv);
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetEvents(XETC *tc, Bool set_flag, EventFlags events)
#else
int XETrapSetEvents(tc, set_flag, events)
    XETC       *tc;
    Bool       set_flag;
    EventFlags events;
#endif
{
    XETCValues tcv;
    int status = True;
    int i;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    XETrapSetCfgFlagEvent(&tcv, valid, True);
    XETrapSetCfgFlagEvent(&tcv, data, set_flag);
    for (i=KeyPress; i<=MotionNotify; i++)
    {
        XETrapSetCfgFlagEvt(&tcv, i, BitValue(events, i));
    }
    status = XEChangeTC(tc, (CARD32)TCEvents, &tcv);
    return(status);
}

#ifdef FUNCTION_PROTOS
int XETrapSetStream(XETC *tc, Bool set_flag, XESTREAM *stream_id,
    CARD8 family, Bool output_flag, Bool input_flag)
#else
int XETrapSetStream(tc, set_flag, stream_id, family, output_flag, input_flag)
    XETC     *tc;
    Bool     set_flag;
    XESTREAM *stream_id;
    CARD8    family;
    Bool     output_flag;
    Bool     input_flag;
#endif
{
    XETCValues tcv;
    CARD8 flags[2L];
    int status = True;
    
    (void)memset((char *)&tcv,0L,sizeof(tcv));
    if (set_flag)
    {
        if (BitIsFalse(XETrapGetTCLFlags(tc,flags), family))
        {   /* Transport not available on Toolkit Side! */
            char *params = XEFamilyIDToString(family);
            int num_params = 1L;
            XtErrorMsg("NoSuchXport", "XESetStream", "XTrapToolkitError",
                "%s not available", &params, &num_params);
            status = False;
        }
        else
        {
            /* 2nd param is conditional statement to select "family"
             * based upon output flag.  Note, XETrapSetCfgTransport is 
             * a macro.  A conditional within a macro! Neat huh?
             */
            XETrapSetCfgTransport(&tcv, output_flag ? family : FamilyXLib);
            XETrapSetCfgFlagTransOut(&tcv, valid, True);
            XETrapSetCfgFlagTransOut(&tcv, data, output_flag);
#ifndef _XINPUT
            /* 2nd param is conditional statement to select "family"
             * based upon input flag.  Note, XETrapSetCfgReadTransport is 
             * a macro.  A conditional within a macro! Neat huh?
             */
            XETrapSetCfgReadTransport(&tcv, input_flag ? family : FamilyXLib);
            XETrapSetCfgFlagTransIn(&tcv, valid, True);
            XETrapSetCfgFlagTransIn(&tcv, data, input_flag);
#endif    
            if (family != FamilyXLib)
            {
                if (stream_id)
                {
                    XETrapSetCfgStream(&tcv, stream_id);
                    XETrapSetCfgAddress(&tcv, stream_id->addr2);
                }
                else
                {   /* Want's streams but no vaild stream! */
                    status = False;
                }
            }
            else
            {
                XETrapSetCfgStream(&tcv, NULL);
            }
        }
    }
    else
    {
        XETrapSetCfgStream(&tcv, NULL);
        XETrapSetCfgTransport(&tcv, FamilyXLib);
        XETrapSetCfgFlagTransOut(&tcv, valid, True);
        XETrapSetCfgFlagTransOut(&tcv, data, False);
#ifndef _XINPUT
        XETrapSetCfgReadTransport(&tcv, FamilyXLib);
        XETrapSetCfgFlagTransIn(&tcv, valid, True);
        XETrapSetCfgFlagTransIn(&tcv, data, False);
#endif
    }
    if (status == True)
    {
        status = XEChangeTC(tc, TCStream, &tcv);
    }
    return(status);
}

#ifdef FUNCTION_PROTOS
CARD8 XEStreamType(CARD8 family)
#else
CARD8 XEStreamType(family)
    CARD8 family;
#endif
{
    CARD8 stream_type = ~0L;
    switch(family)
    {
        case FamilyInternet:
            stream_type = StreamInternet;
            break;
        case FamilyDECnet:
            stream_type = StreamDECnet;
            break;
        case FamilyMbx:
            stream_type = StreamMbx;
            break;
        case FamilyLAT:
            stream_type = StreamLAT;
            break;
        default:
            break;
    }
    return(stream_type);
}

#ifdef FUNCTION_PROTOS
Bool XESetCmdGateState(XETC *tc, CARD8 type, Bool *gate_closed,
    CARD8 *next_key, Bool *key_ignore)
#else
Bool XESetCmdGateState(tc, type, gate_closed, next_key, key_ignore)
    XETC    *tc;
    CARD8   type;
    Bool    *gate_closed;
    CARD8   *next_key;
    Bool    *key_ignore;
#endif
{
    
    *key_ignore = False;
    if (XETrapGetTCFlagCmdKeyMod(tc,data) == True)
    {
        switch (type) 
        {
            case KeyPress:
                if (*next_key == XEKeyIsEcho)
                {
                    break;
                }
                *gate_closed = True;
                *next_key = XEKeyIsClear;
                break;
    
            case KeyRelease:
                if (*next_key == XEKeyIsEcho)
                {
                    *next_key = XEKeyIsClear;
                    break;
                }
                if (*next_key == XEKeyIsClear)
                {
                    *next_key = XEKeyIsEcho;
                }
                else
                {   /* it's XEKeyIsOther, so Clear it */
                    *next_key = XEKeyIsClear;
                }
                *gate_closed = False;
                *key_ignore = True;
                break;
    
            default: break;
        }
    }
    else
    {
        switch (type)
        {
            case KeyPress:
                if (*next_key == XEKeyIsEcho)
                {
                    *gate_closed = False;
                    break;
                }
                /* Open gate on cmd key release */
                if ((*next_key == XEKeyIsOther) && 
                    *gate_closed == True)
                {
                    break;
                }
                *gate_closed = True;
                *next_key = XEKeyIsClear;
                break;

            case KeyRelease:
                if (*next_key == XEKeyIsClear)
                {
                    *next_key = XEKeyIsEcho;
                    break;
                }
    
                if (*next_key == XEKeyIsEcho)
                {
                    *next_key = XEKeyIsClear;
                    break;
                }
    
                *gate_closed = False;
                *key_ignore = True;
                *next_key = XEKeyIsClear;
                break;
    
            default: 
                break;
        }
    }

    return(*gate_closed);
}
