/*****************************************************************************
Copyright 1987, 1988, 1989, 1990 by Digital Equipment Corporation, Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *
 *  ABSTRACT:
 *
 *      Provides routines that simulate DECWindows user input by inserting 
 *      input events into the DECwindows server input event queue.
 *      This module is *very* non-portable!
 */
#ifndef lint
static char SCCSID[] = "@(#)xtrapddvmssim.c	1.7 - 90/11/15  ";
#endif
#ifdef vms

/*-----------------*
 *  INCLUDE FILES  *
 *-----------------*/

#pragma builtins        /* Allow VMS-specific builtins for VAX instructions */

#include "stdio.h"
#include "descrip.h"
#include "ssdef.h"
#include "secdef.h"
#include "psldef.h"
#include "iodef.h"
#include "lnmdef.h"                    /* Logical name definitions */
#include "X.h"                     /* X event definitions */
#include "Xproto.h"                /* X event definitions */
#include "decwdef.h"                   /* X server definitions */
#include "decwcommon.h"                /* X server definitions */

#ifndef XTrapExtName
#define XTrapExtName "DEC-XTRAP"
#endif
/*----------*
 *  MACROS  *
 *----------*/

/*  Check status of procedure calls - On bad status, 
 *  write a message to SYS$ERROR and return
 *  the status code to the caller (we don't want to 
 *  signal within the server process).
 */
#define     $check_status(errstr)               \
    if ( (status&1) == 0L )                      \
        {                                       \
        ErrorF("%s:  %s error: %s\n", XTrapExtName, progname, (errstr));  \
        ErrorF("%s:  VMS error code: %x\n", XTrapExtName, status);        \
        return (status);                        \
        }
#define PRT$C_NA 0L	/* no vaxc supplied prtdef so just define prt$c_na */
#define INSQ_INTERLOCK_FAIL     1       /* Failure: Memory interlock failed */
#define REMQ_INTERLOCK_FAIL     1       /* Failure: Memory interlock failed */
#define REMQ_EMPTY              3       /* Failure: No entry to remove */
#define MAX_RETRY              10
#define $INSQTI(rec, que)                                                     \
{                                                                             \
    int i = 0;                                                                \
    for ( ;; )                                                                \
    {                                                                         \
        if (_INSQTI ((rec), (que)) != INSQ_INTERLOCK_FAIL)                    \
            break;                                                            \
        if (++i >= MAX_RETRY)                                                 \
        {                                                                     \
          ErrorF("%s:  Can't insert SimInput entry on queue\n",XTrapExtName); \
            return;                                                           \
        }                                                                     \
    }                                                                         \
}
#define $REMQHI(que, rec)                                                     \
{                                                                             \
    int i = 0, status;                                                        \
    for ( ;; )                                                                \
    {                                                                         \
        if ( (status = _REMQHI ((que), (rec))) != REMQ_INTERLOCK_FAIL )       \
            break;                                                            \
        if (++i >= MAX_RETRY)                                                 \
        {                                                                     \
            ErrorF("%s:  Can't remove SimInput queue entry\n",XTrapExtName);   \
            return;                                                           \
        }                                                                     \
    }                                                                         \
    if (status == REMQ_EMPTY)                                                 \
        *rec = -1;                                                            \
}
#define $REMQTI(que, rec)                                                     \
{                                                                             \
    int i = 0, status;                                                        \
    for ( ;; )                                                                \
    {                                                                         \
        if ( (status = _REMQTI ((que), (rec))) != REMQ_INTERLOCK_FAIL )       \
            break;                                                            \
        if (++i >= MAX_RETRY)                                                 \
        {                                                                     \
            ErrorF("%s:  Unable to remove queue entry\n",XTrapExtName); \
            return;                                                           \
        }                                                                     \
    }                                                                         \
    if (status == REMQ_EMPTY)                                                 \
        *rec = -1;                                                            \
}


/*---------------------*
 *  EXTERNAL ROUTINES  *
 *---------------------*/

static int		SimMoveAstprm;
unsigned long int SYS$CRMPSC();
unsigned long int SYS$SETPRT();

/*--------------------*
 *  FORWARD ROUTINES  *
 *--------------------*/
int sys$assign();
int sys$get_ef();
int sys$qiow();
struct inp_struct *get_last_event_packet();

/*
 * Support for SimMove
 */
int lib$addx();
int lib$emul();
int sys$gettim();
int sys$setimr();
    
/*---------------------*
 *  TYPE DECLARATIONS  *
 *---------------------*/

typedef int quadword[2];

typedef struct _itemlist
    {
    short   bufferLen;
    short   itemCode;
    char    *bufferAddress;
    int	    *returnLength;
    } ITEMLIST;
/*	From xtrap_mapbuffer.h	*/
typedef struct		_BufferDescrRec {
			    unsigned long int	buf_addr;
			    unsigned long int	start_addr;
			    unsigned long int	end_addr;
			 } BufferDescrRec;
typedef struct _BufferDescrRec *BufferDescrPtr;
             

/*---------------------*
 *  DATA DECLARATIONS  *
 *---------------------*/
                                                                            
static short int    output_chan;
static short int 	kb_chan;

/* Just changed this declaration to static. */
/* may need to be allocated in a psect to work */
static struct inb_struct 	*input_headers;	/* shared memory buffer info */

static struct inb_struct *input_queue;	/* input queue header */
static struct inb_struct *free_queue;	/* free list of packets */

static int		SimCursorActive,
                        SimCompressMotion;
static int              sim_not_initialized = TRUE;
static short int	SimXPos;                       
static short int	SimYPos;
static short int	SimMoveLimit;
static short int	SimMoveCount;
static short int	*SimMoveX;
static short int	*SimMoveY;
static quadword		*SimMoveT;
static int		(*SimMoveAstadr)();
static char		*progname    = "SimInit()";
static char             *trnlnm1_err = 
                        "DECW$SERVER0_SCREENS did not translate to anything\n";
static char		*trnlnm2_err = 
                        "DECW$KEYBOARD did not translate to anything\n";
static char		*scrn_err    = 
                        "Cannot assign channel to DECW$SERVER0_SCREENS\n";
static char		*kbd_err     = 
                        "Could not assign channel to DECW$KEYBOARD\n";
static char		*inpbuf_err  = "Error on the QIO to enable INPUT\n";
static char _align (LONGWORD)
                *undefined  = "???";    /* String for undefined type values */

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimInit -- opens a channel to the workstation screen for
**	cursor warping, and PFNMAPs the server's input event queue.
**--
**/
int
SimInit()
{
    int iosb[2],
        ast_block[3],
        status,
        num_pfns,
        index,
	temp,
	len,
        *pPfns;

    ITEMLIST trnlnm_itemlist[3];

    $DESCRIPTOR(server_table, "DECW$SERVER0_TABLE");
    $DESCRIPTOR(screen, "DECW$SERVER_SCREENS" );
    $DESCRIPTOR(keyboard, "DECW$KEYBOARD" );

    $DESCRIPTOR( SCREEN,   "          " );
    $DESCRIPTOR( KEYBOARD, "          " );

    index = 0;

    /* Prepare for the translation */
    trnlnm_itemlist[0].bufferLen     = sizeof(index);
    trnlnm_itemlist[0].itemCode      = LNM$_INDEX;
    trnlnm_itemlist[0].bufferAddress = &index;
    trnlnm_itemlist[0].returnLength  = &temp;

    /* Now ready the buffer to receive the translation */
    trnlnm_itemlist[1].bufferLen     = SCREEN.dsc$w_length;
    trnlnm_itemlist[1].itemCode      = LNM$_STRING;
    trnlnm_itemlist[1].bufferAddress = SCREEN.dsc$a_pointer;
    trnlnm_itemlist[1].returnLength  = &SCREEN.dsc$w_length;

    /* Zero the last longword of the structure */
    trnlnm_itemlist[2].itemCode = trnlnm_itemlist[2].bufferLen = 0;

    /* Translate the DECW$SERVER_SCREENS logical name */
    status = SYS$TRNLNM(0,&server_table,&screen,0,&trnlnm_itemlist[0]);
    $check_status(trnlnm1_err);

    trnlnm_itemlist[1].bufferLen     = KEYBOARD.dsc$w_length;
    trnlnm_itemlist[1].bufferAddress = KEYBOARD.dsc$a_pointer;
    trnlnm_itemlist[1].returnLength  = &KEYBOARD.dsc$w_length;
    status = SYS$TRNLNM(0,&server_table,&keyboard,0,&trnlnm_itemlist[0]);
    $check_status(trnlnm2_err);

/*
 *  Allocate a channel for the output device (needed when warping the cursor)
 */
    status = sys$assign( &SCREEN, &output_chan, 0, 0 );
    $check_status(scrn_err);

/* 
 *  Get the info about the shared input queue from the driver
 */
    status = sys$assign( &KEYBOARD, &kb_chan, 0, 0 );
    $check_status(kbd_err);

    num_pfns = 10;   

    status = sys$qiow (0, kb_chan, IO$_SETMODE, &iosb, 0, 0, 
                          IO$K_DECW_ENABLE_INPUT, 
			  &num_pfns, &pPfns, 0, 0, 0);
    $check_status(inpbuf_err);
    
    
    input_headers = XTrap_MapBuffer(num_pfns,pPfns,1);
    if (input_headers->inb$w_version != INB$K_VERSION) {
	    ErrorF("%s:  SIMINIT Incompatible input buffer\n",
                XTrapExtName);
	    return(0);  
    }
    input_queue = input_headers;
    free_queue = &(input_headers->inb$l_free_queue_flink);

/*
 *  set the cursor active by default, compress motion by default
 */
    SimCursorActive = TRUE;
    SimCompressMotion = TRUE;
    sim_not_initialized = FALSE;
    return (1);
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimActivateCursor -- sets whether the cursor is warped when
**	an event is inserted into the server's input event queue.
**--        
**/
void
SimActivateCursor( cursormode )
    int cursormode;  
{
    SimCursorActive = cursormode;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimMotionCompress -- sets whether the pointer motion compression is implemented
**      when MotionNotify events are inserted into the server's input event queue.
**--        
**/
void
SimMotionCompress( compressmode )
    int compressmode;  
{
    SimCompressMotion = compressmode;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimInput -- Warps the cursor (conditionally) and inserts an event 
**	into the server's input event queue.
**--
**/

int
SimInputNew( type, detail, xpos, ypos, screen )
    char type;  
    char detail;
    short int xpos, ypos;
    char screen;        /* Assumes it's valid!  Must check it before call */
{
    unsigned short int iosb[4];
    unsigned int status;
    struct inp_struct *input;	/* Xevent w/flink & blink */

    if (sim_not_initialized)
    {
        SimInit();
    }

    input = (struct inp_struct *) -1;                   /* Set default xEvent address */

    if (type == MotionNotify  &&  SimCompressMotion)
        /*
         *  If the latest xEvent on the input queue is also a MotionNotify, dequeue it, and reuse
         *  it for this MotionNotify.  This implements motion compression.
         */
        input = get_last_event_packet (MotionNotify);   /* Returns xEvent address or -1 */

    if ( input == -1 ) {                                /* If not reusing, dequeue a free buffer */
         $REMQHI (free_queue, &input);
         if (input == -1) {
             ErrorF("%s:  SimInput Couldn't REMQHI an xEvent buffer\n",
                XTrapExtName);
             return FALSE;
         }
    }

    /*
     * If a mouse event, warp the cursor.
     */
    if ( SimCursorActive  &&  (type == MotionNotify || type == ButtonPress || type == ButtonRelease) ) {
        status = sys$qiow (0, output_chan, IO$_SETMODE, iosb, 0, 0,
                            IO$K_DECW_CURSOR_POSITION, xpos, ypos, 0, 0, 0 );
        if ( status != SS$_NORMAL  ||  iosb[0] != SS$_NORMAL )  /* Issue warning, but still queue xEvent */
            ErrorF("%s:  SimInput Cursor warp failed, sts=%%x%x, iosbsts=%%x%x, x=%d, y=%d\n",
                        XTrapExtName, status, iosb[0], xpos, ypos);
    }

    /*
     * Fill the xEvent with data and queue it to the server input event queue.
     */
    input->inp$b_type = type;
    input->inp$b_detail = detail;
    /* Note:  For some reason, screen 0 is represented as "1" in the code.
     * The specific reference to the code is the DIXVMS server facility
     * in VMS_INIT.C
     */
    input->inp$l_root_win = screen + 1L;
    input->inp$w_root_x = SimXPos = xpos;
    input->inp$w_root_y = SimYPos = ypos;
    input->inp$l_timestamp = input_headers->inb$l_timestamp_ms;

    $INSQTI (input, input_queue);
    input_headers->inb$r_sched_flags_overlay.inb$r_flag_bits.inb$v_input_pending = TRUE;

    return TRUE;
}
int SimInput( type, detail, xpos, ypos )
    char type;  
    char detail;
    short int xpos, ypos;
{   /* Compatibility function to call with screen 0 */
    return(SimInputNew(type,detail,xpos,ypos,0L));
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**                
**      SimKeyPress -- simulates the user pressing a key on the keyboard.
**--
**/
int
SimKeyPress( keycode )
    char keycode;
{
    return ( SimInput( KeyPress, keycode, SimXPos, SimYPos ) ) ;
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimKeyRelease -- simulates the user releasing a key on the keyboard.
**--
**/
int
SimKeyRelease( keycode )
    char keycode;
{
    return ( SimInput( KeyRelease, keycode, SimXPos, SimYPos ) ) ;
}                     


/*
**++
**  FUNCTIONAL DESCRIPTION:
**                
**      SimButtonPress -- simulates the user pressing a button on the
**	pointer device (like a mouse).
**--
**/
int
SimButtonPress( buttoncode, newx, newy )
    char buttoncode;
    short int newx; 
    short int newy; 
{
    return ( SimInput( ButtonPress, buttoncode, newx, newy ) ) ;
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimButtonRelease -- simulates the user releasing a button on the
**	pointer device (like a mouse).
**--
**/
int
SimButtonRelease( buttoncode, newx, newy )
    char buttoncode;
    short int newx; 
    short int newy; 
{
    return ( SimInput( ButtonRelease, buttoncode, newx, newy ) ) ;
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimMotion -- simulates the user moving the pointer to a 
**	specified location.
**--
**/
int
SimMotion( newx, newy )
    short int newx; 
    short int newy; 
{
    return ( SimInput( MotionNotify, 0, newx, newy ) ) ;
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimMoveAST --       performs the move
**      SimSubmitMoveAST -- schedules the next move
**--
**/
int 
SimMoveAST()
{
    /* TEMP -- should check status after each */
    SimInput( MotionNotify, 0, 
              SimMoveX[SimMoveCount], 
              SimMoveY[SimMoveCount] );
    SimSubmitMoveAST();
}

int 
SimSubmitMoveAST()
{
    SimMoveCount++;
    if ( SimMoveCount < SimMoveLimit ) {
	sys$setimr( 0, &SimMoveT[SimMoveCount], &SimMoveAST, 0 );  /*TEMP*/
    }
    else {
	free( SimMoveX );
	free( SimMoveY );
	free( SimMoveT );
	if ( SimMoveAstadr != NULL) {
	    (*SimMoveAstadr)(SimMoveAstprm);
	}
    } 
}


/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      SimMove -- simulates the user moving the pointer over a period of time.
**--
**/
int
SimMove( n, x, y, t, astadr, astprm )
    short int n;	/* number of moves */ 
    short int *x;       /* array of x positions */
    short int *y;       /* array of y positions */
    short int *t;       /* array of milliseconds to delay before each move */
    int (*astadr)();	/* function to call when all moves are completed */
    int astprm;         /* parameter to pass when calling astadr */
{

    int i,j,k;
    int longt;
    quadword incrtime;
    int arraysize = n * sizeof( short int );
    int addend = 0;
    int msec = 10000;			/* 10k * 100 nanosec = 1 millisec */

    /* 
     * make a permanent copy of parameters 
     */
    SimMoveLimit = n;
    SimMoveX = (short int *) malloc( arraysize );
    memcpy( SimMoveX, x, arraysize );
    SimMoveY = (short int *) malloc( arraysize );
    memcpy( SimMoveY, y, arraysize );
    SimMoveAstadr = astadr;
    SimMoveAstprm = astprm;

    /* 
     * do the rest of the initialization 
     */
    SimMoveCount = -1;

    SimMoveT = (quadword *) malloc( n * sizeof( quadword ) );
    sys$gettim( &SimMoveT[0] );

    longt = (int) t[0];
    lib$emul( &longt, &msec, &addend, &incrtime );
    lib$addx( &incrtime, &SimMoveT[0], &SimMoveT[0], 0 );

    for ( i=1; i<n; i++ ) {
	longt = (int) t[i];
	lib$emul( &longt, &msec, &addend, &incrtime );
        lib$addx( &incrtime, &SimMoveT[i-1], &SimMoveT[i], 0 );
    }

    /* 
     * rest of the work is done at AST level 
     */
    SimSubmitMoveAST(); 
    return TRUE;
}

/*
**++
**  FUNCTIONAL DESCRIPTION:
**
**      get_last_event_packet -- Dequeues xEvent from the tail of the server input event queue
**      if it matches the passed event_type.  This is for motion compression.
**      Returns xEvent address or -1 if no xEvent was dequeued.
**
**      Swiped from DECWMACS.MAR; also see IMSUB.MAR (subroutine IM$$MOUSE_EVENTS).
**--
**/
struct inp_struct
*get_last_event_packet (event_type)
    char event_type;
{
    struct inp_struct *input;           /* Xevent w/flink & blink */

    $REMQTI (input_queue, &input);                      /* Dequeue from tail */
    if ( input != -1 )                                  /* If we got one */
        if ( input->inp$b_type != event_type ) {        /*  but it doesn't match the event_type */
            $INSQTI (input, input_queue);                /* then put it back on the queue */
            return -1;
        }

    return input;
}

/*
*==============================================================================
*
* buffer_address = XTrap_MapBuffer(nPfns, pPfns, ncontig)
*
* Map non-paged buffer shared by the X11 server and a driver
* into the address space of the server.
*
* RETURNS:
*
*   buffer_address = address of the buffer in server process address space
*
* ARGUMENTS:
*
*	nPfns = number of pages to map
*	pPfns = pointer to arrays of pfns. ncontig pages per pfn.
*	ncontig = number of pages per pPfn array entry. Usually 1.
*
*==============================================================================
*/

int
XTrap_MapBuffer(nPfns, pPfns, ncontig)
	int		nPfns;		/* # pfns in buffer */
	int		*pPfns;
	int		ncontig;
{
	unsigned long int	*pPfn;		/* pointer to the current pfn */
	unsigned long int	buffer_address;	/* address of buffer region */
	unsigned long int	buffer_size;
	BufferDescrPtr		pBuf;
	unsigned short int	iosb[4];	/* i/o status block */
	unsigned long int	cond_value;	/* VMS condition value */
	unsigned long int	pInadr[2];	/* region addresses wanted */
	unsigned long int	pRetadr[2];	/* region addresses got */
	int			i;		/* counter for pfn mapping */
	/*
	* Get enough pages to map over the buffer and then some.
	*/
	buffer_size = (nPfns + 2)*512 + 511 + sizeof(BufferDescrRec);
	buffer_address = (unsigned long int) malloc(buffer_size);
	pInadr[0] = ((buffer_address + sizeof(BufferDescrRec)) & ~(0x1FF)) + 512;
	pBuf = (BufferDescrPtr) (pInadr[0] - sizeof(BufferDescrRec));
	pBuf->buf_addr = buffer_address;
	pBuf->start_addr = pInadr[0];


	/*
	* Make first page a no access page
	*/
	pInadr[1] = pInadr[0] + 511;
	cond_value = SYS$SETPRT(pInadr, 0, 0, PRT$C_NA, 0);
	if (!(cond_value & SS$_NORMAL))
	{
	    ErrorF("%s:  MapBuffer can't set page protection\n",XTrapExtName);
/*	    exit(cond_value); */
	    return(0);
	}

	pInadr[0] = pInadr[0] + 512;
	buffer_address = pInadr[0];
	pInadr[1] = pInadr[0] + 512*ncontig - 1;

	pPfn = pPfns;
	for (i = 0; i < nPfns; i += ncontig)
	{
	    cond_value = SYS$CRMPSC(pInadr, pRetadr, PSL$C_USER,
		(SEC$M_PFNMAP | SEC$M_WRT), 0, 0, 0, 0, ncontig, *pPfn, 0, 0);
	    if (!(cond_value & SS$_NORMAL))
	    {
		ErrorF("%s:  MapBuffer can't map page\n",XTrapExtName);
/*		exit(cond_value); */
		return(0);
	    }

	    if ((pInadr[0] != pRetadr[0]) || (pInadr[1] != pRetadr[1]))
	    {
		ErrorF("%s:  MapBuffer mapping error\n",XTrapExtName);
		return(0);
	    }
	    pPfn++;
	    pInadr[0] += 512*ncontig;
	    pInadr[1] += 512*ncontig;
	}
	/*
	* Make page following the buffer no access
	*/
	pInadr[1] = pInadr[0] + 511;
	pBuf->end_addr = pInadr[1]; /* end of cretva free area */
	cond_value = SYS$SETPRT(pInadr, 0, 0, PRT$C_NA, 0);
	if (!(cond_value & SS$_NORMAL))
	{
	    ErrorF("%s:  MapBuffer can't set page protection\n",XTrapExtName);
/*	    exit(cond_value); */
	    return(0);
	}
	return(buffer_address);
}
#endif /* vms */
